/*
 * test_elsa.c - Check if ELSA can be used by the running system.
 *
 * Copyright (C) 2005 BULL SA.
 * Written by Guillaume Thouvenin <guillaume.thouvenin@bull.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>


#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>

#define acctfile "/tmp/bsdacct.tmp"

/**
 * is_root - Check if you are root
 *
 * It returns 1 if your are root and 0 otherwise
 */
int is_root()
{
	return (getuid() == 0) ? 1 : 0;
}

/**
 * has_bsdacct_v3 - Check if the running system supports
 *                 BSD Process Accounting version 3
 *
 * It returns 1 if it supports BSD v3 and 0 otherwise
 */
int has_bsdacct_v3()
{
	int retval = 0;
	int acct_fd;
	int status;
	pid_t pid;
	FILE *acct_fp;
	char buffer[2];

	/* Start process accounting */
	/* We create a file to record accounting data */
	if ((acct_fd = creat(acctfile, S_IRUSR | S_IWUSR)) == -1) {
		perror("Cannot create accouting file");
		return 0;
	}
	close(acct_fd);

	if (acct(acctfile) == -1) {
		perror("Cannot start process accounting");
		goto fini;
	}

	/* To test the version of the BSD accounting we create a child 
	 * process and we check the ac_version field in the accounting 
	 * file. The goal is to have a non-empty file.
	 */
	pid = fork();
	switch (pid) {
	case 0:
		/* Child */
		/* We run a command to fill the accounting file */
		execv("/bin/sync", NULL);
	case -1:
		/* Error */
		perror("Cannot fork");
		retval = -1;
		goto fini;
	default:
		/* parent */
		waitpid(pid, &status, 0);
	}

	/* Stop process accouting */
	acct(NULL);

	/* Now we can read the accounting file and look for the version */
	if ((acct_fp = fopen(acctfile, "r")) == NULL) {
		perror("Cannot read accounting data");
		return 0;
	}

	if (fread(buffer, sizeof(char), sizeof(buffer), acct_fp) !=
	    sizeof(buffer)) {
		perror("Cannot read the first two items");
		fclose(acct_fp);
		goto fini;
	} else 
		retval = ((int)buffer[1] == 3)?1:0;

	/* Close the stream and delete the file */
	fclose(acct_fp);
fini:
	unlink(acctfile);
	return retval;
}

/**
 * main - The entry point
 *
 * Tests done are:
 * 	1) Check if you have root privileges
 * 	2) 
 */

int main()
{
	if (!is_root()) {
		printf("You must be root to run this program\n");
		return 0;
	}

	if (!has_bsdacct_v3()) {
		printf("It seems that your system is not compiled with \n");
		printf("BSD Process Accounting version 3 support.\n");
		return 0;
	}

	return 0;
}
